#version 430 core

#include "global.inc"

in VertexOutput
{
	vec2 uv;
} vertexOutput;

uniform sampler2D lightedScene;
uniform sampler2D bloom01;
uniform sampler2D bloom02;

out vec4 fragColor;

// Adapted: http://callumhay.blogspot.com/2010/09/gaussian-blur-shader-glsl.html
vec4 gaussianBlur(int kernelSize, float sigma, vec2 step, sampler2D texSampler, vec2 texCoord)
{
    int kernelHalfSize = kernelSize / 2;

    // Incremental Computation of the Gaussian
    // https://developer.nvidia.com/gpugems/GPUGems3/gpugems3_ch40.html
    vec3 gaussCoeff;
    gaussCoeff.x = 1.0 / (sqrt(2.0 * pi) * sigma);
    gaussCoeff.y = exp(-0.5 / (sigma * sigma));
    gaussCoeff.z = gaussCoeff.y * gaussCoeff.y;

    vec4 avgValue = vec4(0.0);
    float gaussCoeffSum = 0.0;

    avgValue += texture2D(texSampler, texCoord) * gaussCoeff.x;
    gaussCoeffSum += gaussCoeff.x;
    gaussCoeff.xy *= gaussCoeff.yz;

    for (int i = 1; i <= kernelHalfSize; i++)
    {
        avgValue += texture2D(texSampler, texCoord - float(i) * step) * gaussCoeff.x;
        avgValue += texture2D(texSampler, texCoord + float(i) * step) * gaussCoeff.x;

        gaussCoeffSum += 2.0 * gaussCoeff.x;
        gaussCoeff.xy *= gaussCoeff.yz;
    }

    return avgValue / gaussCoeffSum;
}

void main() 
{
	vec2 textureSizeInv = 1.0 / vec2(viewport.screenWidth, viewport.screenHeight);

	#ifdef HIGH_PASS
		vec3 color = texture(lightedScene, vertexOutput.uv).rgb;
		
		// additional median filter to remove single bright pixels (scene was flickering heavily)
		color = median(lightedScene, vertexOutput.uv, textureSizeInv);
		
		// Use a multiplication factor for the whole image: high values are still high, low values ar still low.
		// This also helps reduce flickering.
		// float factor = 0.3;
		// fragColor = vec4(color * factor, 1.0);

		float lum = luminanceFromColor(color);
		fragColor = vec4(color * max(lum - 1.0, 0), 1.0);

	#endif

	#ifdef BLUR_HORIZONTAL
		vec2 step = vec2(textureSizeInv.x, 0);
		fragColor = gaussianBlur(25, 3.5, step, bloom01, vertexOutput.uv);
	#endif
		
	#ifdef BLUR_VERTICAL
		vec2 step = vec2(0, textureSizeInv.y);
		fragColor = gaussianBlur(25, 5.5, step, bloom02, vertexOutput.uv);
	#endif

	#ifdef COMBINE
		float bloomFactor = 0.05;
		fragColor = vec4(texture(bloom01, vertexOutput.uv).rgb * bloomFactor, 1.0);
		fragColor += vec4(texture(lightedScene, vertexOutput.uv).rgb, 1);
	#endif
}
