#version 430 core

#include "global.inc"

// ----------------------------------
// default shader for everything
// 
// Available defines:
// - DIFFUSE
// - NORMAL
// - TRANSPARENT | CUTOUT
// - GBUFFER
// ----------------------------------

in VertexOutput {
	vec3 positionWorld;
	vec3 normal;

	#ifdef NORMAL
		vec3 tangent;
		vec3 bitangent;
		mat3 tangentSpace;
	#endif

	vec2 uv0;

	#ifdef LIGHTMAP
		vec2 uv1;
	#endif

	#ifdef SHADOW
		vec4 lightSpacePos[NUM_SHADOWMAPS];
	#endif

	#ifdef SHADOWMAP
	    vec4 projPos;
	#endif

} vertexOutput;


// ----------------------------------
// Uniforms
// ----------------------------------

uniform vec4 albedo;	// Main color. Alpha parameter controls transparency (if TRANSPARENT) or cutout level (if CUTOUT)
uniform float metallic;
uniform float roughness;

uniform vec2 texTiling = vec2(1.0, 1.0);

#ifdef DIFFUSE
	uniform sampler2D diffuseTex;
#endif

#ifdef NORMAL
	uniform sampler2D normalTex;
	uniform float normalStrength = 0;
#endif

#ifdef METAL_ROUGHNESS
	uniform sampler2D metalRoughnessTex;
#endif

#ifdef TEST_CUBE
	uniform samplerCube cubeTex;
#endif

#ifdef SHADOWMAP
    uniform sampler2D shadowMap;
#endif


// ----------------------------------
// Local values
// ----------------------------------

#ifdef GBUFFER
    layout (location = 0) out vec3 gPosition;
    layout (location = 1) out vec3 gNormal;
    layout (location = 2) out vec3 gAlbedo;
    // r-channel: metallic, g-channel: roughness, b-channel: currently unused (maybe material id?)
    layout (location = 3) out vec3 gMetallicRoughness;

#else
    out vec4 fragColor;
#endif

void main() 
{
	Surface surf;
	surf.albedo = albedo.rgb;
	surf.position = vertexOutput.positionWorld.xyz;
	surf.view = normalize(camera.position - surf.position);

	vec2 uv0 = vertexOutput.uv0 * texTiling;

	#ifdef METAL_ROUGHNESS
		vec4 metalRoughness = texture(metalRoughnessTex, uv0);
		surf.metallic = metalRoughness.r;
		surf.roughness = metalRoughness.a;
	#else	
		surf.metallic = metallic;
		surf.roughness = roughness;
	#endif

	#ifdef SHADOW
		surf.lightSpacePos = vertexOutput.lightSpacePos;
	#endif

	#ifdef LIGHTMAP
		surf.lightMapUv = vertexOutput.uv1;
	#endif

	float alpha = 1.0;
	#ifdef TRANSPARENT
		alpha = albedo.a;
	#endif

	#ifdef DIFFUSE
		vec4 diffuse = texture(diffuseTex, uv0);
		surf.albedo *= diffuse.rgb;

		#ifdef CUTOUT
			if (diffuse.a < albedo.a)
				discard;
		#endif

		#ifdef TRANSPARENT
			alpha *= diffuse.a;
		#endif
	#endif

	vec3 normal = normalize(vertexOutput.normal);
	#ifdef NORMAL
		normal = texture(normalTex, uv0).rgb;
		normal = normal * 2.0f - 1.0f;   
		normal.xy *= normalStrength;
		normal = normalize(vertexOutput.tangentSpace * normalize(normal)); 
	#endif
	surf.normal = normal;

    #ifdef GBUFFER
        gPosition = surf.position;
        gNormal = surf.normal;
        // gViewDir = surf.view;
        gAlbedo = surf.albedo;
        gMetallicRoughness = vec3(surf.metallic, surf.roughness, 1.0);
        #ifdef SHADOWMAP
            float lightDepth = textureProj(shadowMap, vertexOutput.projPos.xyw).r;
            if (lightDepth < (vertexOutput.projPos.z-0.005)/vertexOutput.projPos.w) {
                gAlbedo *= 0.5;
            }
        #endif
    #else
    	// forward pass is only used for updating cubemaps
    	// use a very basic illum model

		float kD = 1.0 - surf.metallic;
		vec3 shaded = kD * surf.albedo * textureLod(skybox, surf.normal, 1024).rgb;

    	#ifdef XX_NUM_LIGHTS
			for (int i = 0; i < NUM_LIGHTS; i++)
			{
				Light light = lights[i];

				vec3 L = light.type == LightTypeDirectional ? -light.direction : light.position - surf.position;
				float distance = length(L);
				L = normalize(L);
				float NdotL = max(dot(surf.normal, L), 0.0);

				vec3 lightColor = light.color * light.intensity * attenuate(light, L, distance);
				shaded += kD * surf.albedo * NdotL * lightColor;

				#ifdef SHADOW
					if (light.castsShadows)
					{
						float shadow = calcShadow(light, surf.lightSpacePos[light.shadowMapIndex]);
						shaded *= vec3(shadow);
					}
				#endif
			}
		#endif	// NUM_LIGHTS

		fragColor = vec4(shaded, alpha);
        //vec3 shaded = shade(surf);

        //#ifdef FOG
        ///float distance = length(posWorld - camera.position);
        // finalColor = calcFog(finalColor, distance);
        //#endif
	#endif
}

