#version 430 core

#include "global.inc"

in VertexOutput 
{
	vec2 uv;
} vertexOutput;

uniform sampler2D gPosition;
uniform sampler2D gNormal;
uniform sampler2D gAlbedo;
uniform sampler2D gMetallicRoughness; // r-channel: metallic, g-channel: roughness, b-channel: currently unused

#ifdef NUM_REFLECTION_PROBES
	uniform samplerCube reflectionProbeTex[NUM_REFLECTION_PROBES];
	uniform ReflectionProbe reflectionProbes[NUM_REFLECTION_PROBES];
#endif

out vec4 fragColor;

#ifdef SSDO
    const float coveredSolidAngle = (2 * pi) / NUM_SAMPLES;

    /** Calculates the directional occlusion for one pixel.
    */
    vec3 directionalOcclusion(vec3 position, vec3 viewSpacePosition, mat3 TBN, vec3 normal, samplerCube environment, float lod) {
        vec3 result = vec3(0);

        for (int i = 0; i < NUM_SAMPLES; ++i) {
            vec3 sampleDirection = TBN * samples[i].xyz;
            vec3 samplePoint = position + sampleDirection * samples[i].w;

            vec4 viewSpaceSample = camera.view * vec4(samplePoint, 1);
            vec2 sampleUV = getSampleUV(viewSpaceSample);
            vec3 viewSpaceOccluder = (camera.view * vec4(texture(gPosition, sampleUV).rgb, 1)).xyz;
            float distance = length(viewSpacePosition - viewSpaceOccluder);
            float visibility = getVisibility(viewSpaceSample.xyz, viewSpaceOccluder, distance);

            float cosTheta = dot(normal, sampleDirection);
            vec3 li = textureLod(environment, sampleDirection, lod).xyz;

            result += li * visibility * cosTheta;
            //result += abs(texture(gNormal, sampleUV).rgb);
            //result += vec3(visibility) / NUM_SAMPLES;
            //result += vec3(distance);
            //result += sampleDirection;
        }
        return result * oneOverPi * coveredSolidAngle;
    }
#endif

void main() 
{
	vec2 uv = vertexOutput.uv;
	
	vec3 position = texture(gPosition, uv).rgb;
	vec3 normal = texture(gNormal, uv).rgb;
    vec3 albedo = texture(gAlbedo, uv).rgb;
    vec3 metallicRoughness = texture(gMetallicRoughness, uv).rgb;		
	float metallic = metallicRoughness.r;
	float roughness = metallicRoughness.g;

	vec3 viewDir = normalize(camera.position - position);
    float NdotV = max(dot(normal, viewDir), 0.0f);
    vec3 R = reflect(-viewDir, normal);

	vec3 F0 = vec3(0.04); // we assume F0 = 0.04 for dielectrics
	F0 = mix(F0, albedo, metallic); // metallic surfaces have a tinted reflection and no diffuse light - use albedo color here
	vec3 fresnel = fresnelSchlickRoughness(NdotV, F0, roughness);
	
	float mipLevel = mipLevelFromRoughness(roughness);

	#ifdef SSDO
	    vec3 viewSpacePosition = (camera.view * vec4(position, 1)).xyz;
        mat3 TBN = getCoordinateSystem(normal);

	    float ssdoMipLevel = 1024;
	    ssdoMipLevel = mipLevel;
	    ssdoMipLevel = 0;
	#endif

	vec3 kD = vec3(1.0) - fresnel;
	kD *= 1.0 - metallic; // no diffuse for metallics

	vec3 diffuse = vec3(0);
	vec3 specular = vec3(0);
	float accu = 0;

	#ifdef NUM_REFLECTION_PROBES
		for (int i = 0; i < NUM_REFLECTION_PROBES; i++)
		{
			ReflectionProbe probe = reflectionProbes[i];
			vec3 localPosition = (probe.worldInverse * vec4(position, 1)).xyz;
			
			if (insideAABB(vec3(-0.5), vec3(0.5), localPosition))
			{
				//float distance = length(localPosition);
				//float weight = 1; // - smoothstep(0.4, 0.5, distance);
				float distToPlane = 0.5 - max(max(abs(localPosition.x), abs(localPosition.y)), abs(localPosition.z));
				float weight = smoothstep(0.0, 0.1, distToPlane);

				// Parallax correction
				vec3 localReflection = (probe.worldInverse * vec4(R, 0)).xyz;

				vec3 unit = vec3(1, 1, 1);
				vec3 intersectPlane1 = (unit - localPosition) / localReflection;
				vec3 intersectPlane2 = (-unit - localPosition) / localReflection;

				vec3 furthestPlane = max(intersectPlane1, intersectPlane2);
				float dist = min(furthestPlane.x, min(furthestPlane.y, furthestPlane.z));
				vec3 intersectPosWorld = position + R * dist;
				vec3 lookup = intersectPosWorld - probe.position;

				#ifdef SSDO
				    diffuse += directionalOcclusion(position, viewSpacePosition, TBN, normal, reflectionProbeTex[i], ssdoMipLevel) * weight;
				#else
				    diffuse += textureLod(reflectionProbeTex[i], normal, 1024).rgb * weight;
				#endif
				specular += textureLod(reflectionProbeTex[i], lookup, mipLevel).rgb * weight;

				accu += weight;
			}
		}

		if (accu > 1.0)
		{
			diffuse /= accu;
			specular /= accu;
			accu = 1.0;
		}
	#endif

	#ifdef SSDO
	    diffuse += directionalOcclusion(position, viewSpacePosition, TBN, normal, skybox, ssdoMipLevel) * (1 - accu);
	    //diffuse = directionalOcclusion(position, viewSpacePosition, TBN, normal, skybox, ssdoMipLevel);
	    //specular = vec3(0);
	#else
	    diffuse += textureLod(skybox, normal, 1024).rgb * (1 - accu);
	#endif
    specular += textureLod(skybox, R, mipLevel).rgb * (1 - accu);

	diffuse = diffuse * kD * albedo;
	specular = fresnel * specular;

	fragColor = vec4(diffuse + specular, 1);
}

