
// --------------------------------------------
// Global params
// --------------------------------------------



// --------------------------------------------
// Uniforms
// --------------------------------------------

uniform struct Frame
{
	float dt;
	float totalTime;
} frame;


uniform struct Camera
{
	float near;
	float far;
	vec3 position;
	mat4 eye;
	mat4 view;
	mat4 projection;
	mat4 viewProjection;
} camera;

uniform struct Viewport
{
	float x;
	float y;
	float width;
	float height;
	int screenWidth;
	int screenHeight;
} viewport;

// Per object matrices
uniform struct Object
{
	mat4 world;
	mat4 worldInverse;
	mat4 normal;
	mat4 worldView;
	mat4 worldViewProjection;
} object;

// set after depth prepass
uniform sampler2D depthTex;

// if a skybox is present
uniform samplerCube skybox;


// --------------------------------------------
// Utils
// --------------------------------------------

const float pi = 3.1415926535897932384626433832795;

float restoreZ(float depth)
{
	// http://web.archive.org/web/20130416194336/http://olivers.posterous.com/linear-depth-in-glsl-for-real
	return 2.0f * camera.near * camera.far / (camera.far + camera.near - (2.0 * depth - 1.0) * (camera.far - camera.near));
}

/*
float mapRange(float qMin, float qMax, float pMin, float pMax, float value)
{
	return (pMax - pMin) * (value - qMin) / (qMax - qMin) + pMin;
}
*/

// map from u to x
float mapRange(float u1, float u2, float x1, float x2, float value)
{
	return value * ((x2 - x1) / (u2 - u1)) + (x1 * u2 - x2 * u1) / (u2 - u1);
}

vec2 mapVec2(vec2 u1, vec2 u2, vec2 x1, vec2 x2, vec2 value)
{
	return value * ((x2 - x1) / (u2 - u1)) + (x1 * u2 - x2 * u1) / (u2 - u1);
}

bool insideAABB(vec3 min, vec3 max, vec3 p)
{
	return p.x >= min.x && p.x <= max.x && p.y >= min.y && p.y <= max.y && p.z >= min.z && p.z <= max.z;
}

float luminanceFromColor(vec3 color)
{
	return dot(color, vec3(0.2126, 0.7152, 0.0722));
}


// Median filter taken from: http://casual-effects.com/research/McGuire2008Median/median.pix

/*
3x3 Median
Morgan McGuire and Kyle Whitson
http://graphics.cs.williams.edu


Copyright (c) Morgan McGuire and Williams College, 2006
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

// Change these 2 defines to change precision,
#define vec vec3
#define toVec(x) x.rgb

//#define vec vec4
//#define toVec(x) x.rgba

#define s2(a, b)				temp = a; a = min(a, b); b = max(temp, b);
#define mn3(a, b, c)			s2(a, b); s2(a, c);
#define mx3(a, b, c)			s2(b, c); s2(a, c);

#define mnmx3(a, b, c)			mx3(a, b, c); s2(a, b);                                   // 3 exchanges
#define mnmx4(a, b, c, d)		s2(a, b); s2(c, d); s2(a, c); s2(b, d);                   // 4 exchanges
#define mnmx5(a, b, c, d, e)	s2(a, b); s2(c, d); mn3(a, c, e); mx3(b, d, e);           // 6 exchanges
#define mnmx6(a, b, c, d, e, f) s2(a, d); s2(b, e); s2(c, f); mn3(a, b, c); mx3(d, e, f); // 7 exchanges

vec3 median(sampler2D texture, vec2 uv, vec2 textureSizeInv) {

  vec v[9];

  // Add the pixels which make up our window to the pixel array.
  for(int dX = -1; dX <= 1; ++dX) {
    for(int dY = -1; dY <= 1; ++dY) {		
      vec2 offset = vec2(float(dX), float(dY));
		    
      // If a pixel in the window is located at (x+dX, y+dY), put it at index (dX + R)(2R + 1) + (dY + R) of the
      // pixel array. This will fill the pixel array, with the top left pixel of the window at pixel[0] and the
      // bottom right pixel of the window at pixel[N-1].
      v[(dX + 1) * 3 + (dY + 1)] = toVec(texture2D(texture, uv + offset * textureSizeInv));
    }
  }

  vec temp;

  // Starting with a subset of size 6, remove the min and max each time
  mnmx6(v[0], v[1], v[2], v[3], v[4], v[5]);
  mnmx5(v[1], v[2], v[3], v[4], v[6]);
  mnmx4(v[2], v[3], v[4], v[7]);
  mnmx3(v[3], v[4], v[8]);
  
  // toVec(gl_FragColor) = v[4];
  return v[4];
}

#include "lighting.inc"