
// --------------------------------------------
// Lighting
// --------------------------------------------

uniform vec3 ambientColor;

// This must match the C++ enum
const int LightTypeDirectional = 0;
const int LightTypePoint = 1;
const int LightTypeSpot = 2;

struct Light 
{
	int type;

	vec3 color;
	vec3 position;
	vec3 direction;
	
	float intensity;
	float constant;
	float linear;
	float quadratic;
	
	float innerAngle;
	float outerAngle;
	
	bool castsShadows;
	float shadowStrength;
	float shadowBias;
	int shadowMapIndex;
};

struct ReflectionProbe
{
	vec3 position;
	mat4 world;
	mat4 worldInverse;
};

struct Surface
{
	vec3 albedo;
	float metallic;
	float roughness;
	vec3 position;
	vec3 normal;
	vec3 view;

	#ifdef NUM_SHADOWMAPS
		vec4 lightSpacePos[NUM_SHADOWMAPS];
	#endif

	#ifdef LIGHTMAP
		vec2 lightMapUv;
	#endif
};





float attenuate(Light light, vec3 L, float distance)
{
	float attenuation = 1.0;

	if (light.type == LightTypePoint)
	{
		attenuation = 1.0 / (light.constant + (light.linear * distance) +  (light.quadratic * pow(distance, 2)));
	} 
	
	if (light.type == LightTypeSpot)
	{
		float cosInner = cos(light.innerAngle);
		float cosOuter = cos(light.outerAngle);
		attenuation *= clamp(smoothstep(cosOuter, cosInner, dot(-L, normalize(light.direction))), 0.0f, 1.0f);
	}

	return attenuation;
}

vec3 fresnelSchlick(float cosTheta, vec3 F0)
{
    return F0 + (1.0 - F0) * pow(1.0 - cosTheta, 5.0);
} 

vec3 fresnelSchlickRoughness(float cosTheta, vec3 F0, float roughness)
{
    return F0 + (max(vec3(1.0 - roughness), F0) - F0) * pow(1.0 - cosTheta, 5.0);
}   

// taken from https://github.com/1upon0/sura-2015/blob/master/builtin_shaders-5.0.2f1/CGIncludes/UnityStandardBRDF.cginc
float roughnessToSpecPower (float roughness)
{
	float m = roughness * roughness * roughness + 1e-4f;	
	float n = (2.0 / m) - 2.0;			// http://jbit.net/%7Esparky/academic/mm_brdf.pdf
	n = max(n, 1e-4f);					// prevent possible cases of pow(0,0), which could happen when roughness is 1.0 and NdotH is zero
	return n;
}

float blinnPhong(float roughness, float NdotH)
{
	float m = roughnessToSpecPower(roughness);

	// normalization factor
	// Real-Time Rendering p. 257 and
	// https://github.com/1upon0/sura-2015/blob/master/builtin_shaders-5.0.2f1/CGIncludes/UnityStandardBRDF.cginc
	float norm = (m + 8.0) / 8.0;	

	return pow(NdotH, m) * norm;
}

float mipLevelFromRoughness(float roughness)
{
	// http://casual-effects.blogspot.com/2011/08/plausible-environment-lighting-in-two.html?m=1
	float glossy = roughnessToSpecPower(roughness);
	float mipLevel = log2(512 * sqrt(3)) - 0.5 * log2(glossy + 1);

	// TODO: wrong glossy calculation
	mipLevel -= 2;
	return mipLevel;
}



#ifdef NUM_SHADOWMAPS

uniform mat4 lightViewProjection[NUM_SHADOWMAPS];
uniform sampler2D shadowMaps[NUM_SHADOWMAPS];

float calcShadow(Light light, vec4 lightSpacePos)
{
	int index = light.shadowMapIndex;

	vec3 projCoords = lightSpacePos.xyz / lightSpacePos.w;
	projCoords = projCoords * 0.5 + 0.5;

	float fragDepth = projCoords.z;

	// fragment behind far plane - ignore
	if (fragDepth >= 1.0)
		return 1.0;

	float bias = light.shadowBias;

	float shadow = 0.0;
	vec2 texelSize = 1.0 / textureSize(shadowMaps[index], 0);


	for(int x = -1; x <= 1; x++)
	{
		for(int y = -1; y <= 1; y++)
		{
		    float shadowDepth = texture(shadowMaps[index], projCoords.xy + vec2(x, y) * texelSize).r;
		    shadow += fragDepth - bias > shadowDepth ? 1.0 : 0.0;        
		}    
	}

	shadow /= 9.0;
	return 1 - shadow * light.shadowStrength;
}

#endif	// NUM_SHADOWMAPS


uniform sampler2D lightMap;


#ifdef NUM_LIGHTS
uniform Light lights[NUM_LIGHTS];
#endif

// xyz - normalized direction, w - distance to point
vec4 lightDirectionForPoint(Light light, vec3 point) {
    vec3 L = light.type == LightTypeDirectional ? -light.direction : light.position - point;
    return vec4(normalize(L), length(L));
}

vec3 lightColorAtDistance(Light light, vec3 L, float distance) {
    return light.color * light.intensity * attenuate(light, L, distance);
}

#ifdef SSAO
vec3 shade(Surface surf, float ssao)
#else
vec3 shade(Surface surf)
#endif
{
	#ifdef NUM_LIGHTS
		vec3 fragColor;

		for (int i = 0; i < NUM_LIGHTS; i++){
		    Light light = lights[i];

            vec4 lightDirection = lightDirectionForPoint(light, surf.position);
            vec3 L = lightDirection.xyz;
            float distance = lightDirection.w;

            vec3 H = normalize(L + surf.view);
            float NdotL = max(dot(surf.normal, L), 0.0);
            float NdotV = max(dot(surf.normal, surf.view), 0.0f);
            float NdotH = max(dot(surf.normal, H), 0.0f);
            float LdotH = max(dot(H, L), 0.0f);
            float HdotV = max(dot(H, surf.view), 0.0f);

            vec3 F0 = vec3(0.04);	// we assume F0 = 0.04 for dielectrics
            F0 = mix(F0, surf.albedo, surf.metallic);	// metallic surfaces have a tinted reflection and no diffuse light - use albedo color here

            // Fresnel is effectively the multiplication factor for reflections ( = kS)
            // I still don't get when to use NdotL, NdotH, LdotH, ...
            // fresnel is very subtle with LdotH, but looks nice
            vec3 fresnel = fresnelSchlick(LdotH, F0);

            vec3 kD = vec3(1.0) - fresnel;
            kD *= 1.0 - surf.metallic;	// no diffuse for metallics

            // bring it all together
            vec3 specular = fresnel * blinnPhong(surf.roughness, NdotH);
            vec3 diffuse = kD * surf.albedo;
            vec3 lightColor = light.color * light.intensity * attenuate(light, L, distance);

            vec3 result = (diffuse + specular) * NdotL * lightColor;

            #ifdef SHADOW
                if (light.castsShadows)
                {
                    float shadow = calcShadow(light, lightViewProjection[light.shadowMapIndex] * vec4(surf.position, 1));
                    result *= vec3(shadow);
                }
            #endif

            fragColor += result;
        }

        #ifdef LIGHTMAP
            vec3 lightMapColor = texture(lightMap, surf.lightMapUv).rgb;
            lightMapColor = max(vec3(0.2), lightMapColor);
            fragColor *= lightMapColor;
        #endif

        #ifdef SSAO
            fragColor += (ambientColor * surf.albedo) * ssao;
        #else
            fragColor += ambientColor * surf.albedo;
        #endif

        return fragColor;
    #else
		return surf.albedo * vec3(0.1, 0.1, 0.1);
	#endif	// NUM_LIGHTS
}






uniform vec3 fogColor;
uniform float fogStart;
uniform float fogRange;

vec3 calcFog(vec3 color, float distance)
{
	float amount = clamp((distance - fogStart) / fogRange, 0.0, 1.0);
	return mix(color, fogColor, amount);
}

// SSDO + indirect bounce common
#ifdef SSDO
    // xyz = normalized direction, w = length (in range [0,rMax])
    uniform vec4 samples[NUM_SAMPLES];
    uniform float rMax;
    uniform float depthBias;

    const float oneOverPi = 1 / pi;

    // rotating sample kernel - https://learnopengl.com/Advanced-Lighting/SSAO
    uniform sampler2D rotationVecTex;
    const vec2 noiseScale = vec2(RES_WIDTH/SQRT_NOISE_DIM, RES_HEIGHT/SQRT_NOISE_DIM);
    mat3 getRotatedCoordinateSystem(vec3 v3, vec2 uv) {
        vec3 rotationVec = texture(rotationVecTex, uv * noiseScale).xyz;
        vec3 v1 = normalize(rotationVec - v3 * dot(rotationVec, v3));
        vec3 v2 = cross(v3, v1);
        return mat3(v1, v2, v3);
    }

    mat3 getCoordinateSystem(vec3 v3) {
        vec3 v1 = normalize(abs(v3.x) >= abs(v3.y) ? vec3(v3.z, 0, -v3.x) : vec3(0, -v3.z, v3.y));
        vec3 v2 = cross(v3, v1);
        return mat3(v1, v2, v3);
    }

    vec2 getSampleUV(vec4 viewSpaceSamplePoint) {
        vec4 result = camera.projection * viewSpaceSamplePoint;
        result.xyz /= result.w;
        return (result.xyz * 0.5 + 0.5).xy;
    }

    float getVisibility(vec3 viewSpaceSample, vec3 viewSpaceOccluder, float distance) {
        return abs(viewSpaceOccluder.z) >= abs(viewSpaceSample.z) + depthBias ? 1.0 : smoothstep(0, 1, distance/rMax);
    }
#endif
