#version 430 core

#include "global.inc"

in VertexOutput {
	vec3 positionWorld;
	vec3 normal;

#ifdef NORMALMAP
	vec3 tangent;
	vec3 bitangent;
	mat3 tangentSpace;
#endif

	vec2 uv0;
} vertexOutput;

uniform struct Material
{
	vec3 albedo;
	float metallic;
	float roughness;
	// float ao;

#ifdef DIFFUSEMAP
	sampler2D diffuse;
#endif

#ifdef NORMALMAP
	sampler2D normal;
#endif

} material;

out vec4 fragColor;

const float PI = 3.14159265359;

// the ratio of reflected light
vec3 fresnelSchlick(float cosTheta, vec3 F0)
{
    return F0 + (1.0 - F0) * pow(1.0 - cosTheta, 5.0);
}  

float DistributionGGX(vec3 N, vec3 H, float roughness)
{
    float a      = roughness*roughness;
    float a2     = a*a;
    float NdotH  = max(dot(N, H), 0.0);
    float NdotH2 = NdotH*NdotH;
	
    float num   = a2;
    float denom = (NdotH2 * (a2 - 1.0) + 1.0);
    denom = PI * denom * denom;
	
    return num / denom;
}

float GeometrySchlickGGX(float NdotV, float roughness)
{
    float r = (roughness + 1.0);
    float k = (r*r) / 8.0;

    float num   = NdotV;
    float denom = NdotV * (1.0 - k) + k;
	
    return num / denom;
}
float GeometrySmith(vec3 N, vec3 V, vec3 L, float roughness)
{
    float NdotV = max(dot(N, V), 0.0);
    float NdotL = max(dot(N, L), 0.0);
    float ggx2  = GeometrySchlickGGX(NdotV, roughness);
    float ggx1  = GeometrySchlickGGX(NdotL, roughness);
	
    return ggx1 * ggx2;
}



void main() 
{
	vec3 posWorld = vertexOutput.positionWorld.xyz;
	vec3 N = normalize(vertexOutput.normal);
	vec3 V = normalize(camera.position - posWorld);

#ifdef NORMALMAP
	N = texture(material.normal, vertexOutput.uv0).rgb;
	N = normalize(N * 2.0f - 1.0f);   
	N = normalize(vertexOutput.tangentSpace * N); 
#endif

	float metallic = material.metallic;
	float roughness = material.roughness;
	vec3 albedo = material.albedo;

#ifdef DIFFUSEMAP
	albedo *= texture(material.diffuse, vertexOutput.uv0).rgb;
#endif

	// we assume F0 = 0.04 for dielectrics,
	vec3 F0 = vec3(0.04); 

	// metallic surfaces have a tinted reflection and no diffuse light - use albedo color here
	F0 = mix(F0, albedo, metallic);	

	vec3 Lo = vec3(0.0);
	for (int i = 0; i < NUM_LIGHTS; i++){

		Light light = lights[i];

		// Light dir
		vec3 L;
		//if (light.type == LightTypeDirectional)
			L = normalize(-light.direction);
		//else 
			//L = normalize(light.position - posWorld);

		// Half vector
    	vec3 H = normalize(V + L);

    	//float distance    = length(lightPositions[i] - WorldPos);
        //float attenuation = 1.0 / (distance * distance);
        vec3 radiance     = light.color;	//[i] * attenuation;     

		
		vec3 F  = fresnelSchlick(max(dot(H, V), 0.0), F0);
		//vec3 F  = fresnelSchlick(0, F0);


		float NDF = DistributionGGX(N, H, roughness);       
		float G = GeometrySmith(N, V, L, roughness);   

		vec3 kS = F;
		vec3 kD = vec3(1.0) - kS;
		kD *= 1.0 - metallic;	

		vec3 numerator    = NDF * G * F;
		float denominator = 4.0 * max(dot(N, V), 0.0) * max(dot(N, L), 0.0);
		vec3 specular     = numerator / max(denominator, 0.001); 
		
    	float NdotL = max(dot(N, L), 0.0);        
    	//Lo += (kD * albedo / PI + specular) * radiance * NdotL;
		Lo += (kD * albedo + specular) * radiance * NdotL;
		//Lo += F;
	}
	
	//vec3 ambient = vec3(0.03) * albedo;	// * ao;
	vec3 color   = Lo;

	vec3 R = reflect(-V, N);
	vec3 refl = texture(skybox, R).rgb;
	color = mix(refl, color, 1);

	//color = color / (color + vec3(1.0));
	//color = pow(color, vec3(1.0/2.2)); 
	
	fragColor = vec4(color, 1.0);
}

