#version 430 core

#include "global.inc"

in VertexOutput
{
	vec2 uv;
} vertexOutput;

uniform sampler2D gPosition;
uniform sampler2D gNormal;
uniform sampler2D gAlbedo;
uniform sampler2D gMetallicRoughness; // r-channel: metallic, g-channel: roughness, b-channel: currently unused

uniform samplerCube reflectionProbeTex[NUM_REFLECTION_PROBES];
uniform ReflectionProbe reflectionProbes[NUM_REFLECTION_PROBES];

layout (location = 0) out vec4 lightingResult;
layout (location = 1) out vec4 diffuseBrdf;
//out vec4 fragColor;

void main() {
	vec2 uv = vertexOutput.uv;

	vec3 position = texture(gPosition, uv).rgb;
	vec3 normal = texture(gNormal, uv).rgb;
    vec3 albedo = texture(gAlbedo, uv).rgb;
    vec3 metallicRoughness = texture(gMetallicRoughness, uv).rgb;
	float metallic = metallicRoughness.r;
	float roughness = metallicRoughness.g;

	vec3 viewDir = normalize(camera.position - position);
    float NdotV = max(dot(normal, viewDir), 0.0f);
    vec3 R = reflect(-viewDir, normal);

	vec3 F0 = vec3(0.04); // we assume F0 = 0.04 for dielectrics
	F0 = mix(F0, albedo, metallic); // metallic surfaces have a tinted reflection and no diffuse light - use albedo color here
	vec3 fresnel = fresnelSchlickRoughness(NdotV, F0, roughness);

	vec3 kD = vec3(1.0) - fresnel;
	kD *= 1.0 - metallic; // no diffuse for metallics

    vec3 diffBRDF = kD * albedo;

    vec3 diffuse = vec3(0);
    vec3 specular = vec3(0);

    float accu = 0;
    float mipLevel = mipLevelFromRoughness(roughness);

    // premature optimisation...
    vec4 homWorldPos = vec4(position, 1);
    vec4 homR = vec4(R, 0);
    vec3 negHalf = vec3(-0.5);
    vec3 posHalf = vec3(0.5);
    vec3 unit = vec3(1, 1, 1);

    for (int i = 0; i < NUM_REFLECTION_PROBES; i++) {
        ReflectionProbe probe = reflectionProbes[i];
        vec3 localPosition = (probe.worldInverse * homWorldPos).xyz;

        if (insideAABB(negHalf, posHalf, localPosition)) {
            float distToPlane = 0.5 - max(max(abs(localPosition.x), abs(localPosition.y)), abs(localPosition.z));
            float weight = smoothstep(0.0, 0.1, distToPlane);

            // Parallax correction
            vec3 localReflection = (probe.worldInverse * homR).xyz;

            vec3 intersectPlane1 = (unit - localPosition) / localReflection;
            vec3 intersectPlane2 = (-unit - localPosition) / localReflection;

            vec3 furthestPlane = max(intersectPlane1, intersectPlane2);
            float dist = min(furthestPlane.x, min(furthestPlane.y, furthestPlane.z));
            vec3 intersectPosWorld = position + R * dist;
            vec3 lookup = intersectPosWorld - probe.position;

            //diffuse += textureLod(reflectionProbeTex[i], normal, 1024).rgb * weight;
            specular += textureLod(reflectionProbeTex[i], lookup, mipLevel).rgb * weight;

            accu += weight;
        }
    }
    if (accu > 1.0) {
        diffuse /= accu;
        specular /= accu;
        accu = 1.0;
    }
    //diffuse += textureLod(skybox, normal, 1024).rgb * (1 - accu);
    specular += textureLod(skybox, R, mipLevel).rgb * (1 - accu);

    vec3 environmentLighting = (diffuse * diffBRDF) + (specular * fresnel);

    vec3 directLighting;
    float totalShadow = 1.0;
    for (int i = 0; i < NUM_LIGHTS; i++){
        Light light = lights[i];
        if (light.castsShadows) {
            float shadow = calcShadow(light, lightViewProjection[light.shadowMapIndex] * homWorldPos);
            totalShadow *= shadow;
        }
    }

	vec3 totalLighting = environmentLighting * max(0.5, totalShadow);

	//fragColor = vec4(totalLighting, 1);
	lightingResult = vec4(totalLighting, 1);
	diffuseBrdf = vec4(diffBRDF, 1);
}

