#version 430 core

#include "global.inc"

in VertexOutput
{
	vec2 uv;
} vertexOutput;

uniform sampler2D gPositionHalfRes;
uniform sampler2D gNormalHalfRes;
uniform sampler2D brdfHalfRes;

out vec4 fragColor;

// VOLUMETRIC
uniform float stepSize;

uniform vec3 mediumAlbedo;

const float PI_TIMES_FOUR = pi * 4;

struct Ray {
  vec3 origin;
  vec3 direction;
  vec3 intersection;
  float tMax;           // intersection = origin + direction * tMax;
  vec3 step;            // one step along the ray
};

Ray createRay(vec3 origin, vec3 intersection) {
    vec3 direction = intersection - origin;

    Ray charles;
    charles.origin = origin;
    charles.intersection = intersection;
    charles.direction = normalize(direction);
    charles.tMax = length(direction);
    charles.step = charles.direction * stepSize;
    return charles;
}

// TODO: this needs the intensity as well
vec3 calculateContribution(float shadow, float lightDistance, float travelledDistance, float denominator) {
    return shadow * mediumAlbedo * denominator * exp(-lightDistance * mediumAlbedo) * exp(-travelledDistance * mediumAlbedo) * stepSize;
}

vec3 directionalLightContribution(Ray charles, Light light) {
    vec3 contribution = vec3(0);

    vec3 L = normalize(-light.direction);
    float distance = 1.0;
    vec3 intensity = light.color * light.intensity;
    float denominator = 1.0 / PI_TIMES_FOUR;

    mat4 viewProjection = lightViewProjection[light.shadowMapIndex];

    vec3 currentStep = charles.origin + charles.step;
    for (float t = stepSize; t < charles.tMax; t += stepSize) {
        float shadow = calcShadow(light, viewProjection * vec4(currentStep, 1));

        contribution += calculateContribution(shadow, distance, t, denominator) * intensity;

        currentStep += charles.step;
    }

    return contribution;
}


// SSDO

const float coveredSolidAngle = (2 * pi) / NUM_SAMPLES;

/** Calculates the directional occlusion for one pixel.
*/
vec3 directionalOcclusion(vec3 position, vec3 viewSpacePosition, mat3 TBN, vec3 normal, samplerCube environment, float lod) {
    vec3 result = vec3(0);

    for (int i = 0; i < NUM_SAMPLES; ++i) {
        vec3 sampleDirection = TBN * samples[i].xyz;
        vec3 samplePoint = position + sampleDirection * samples[i].w;

        vec4 viewSpaceSample = camera.view * vec4(samplePoint, 1);
        vec2 sampleUV = getSampleUV(viewSpaceSample);
        vec3 viewSpaceOccluder = (camera.view * vec4(texture(gPositionHalfRes, sampleUV).rgb, 1)).xyz;
        float distance = length(viewSpacePosition - viewSpaceOccluder);
        float visibility = getVisibility(viewSpaceSample.xyz, viewSpaceOccluder, distance);

        float cosTheta = dot(normal, sampleDirection);
        vec3 li = textureLod(environment, sampleDirection, lod).xyz;

        result += li * visibility * cosTheta;
        //result += abs(texture(gNormal, sampleUV).rgb);
        //result += vec3(1 - visibility);
        //result += vec3(distance);
        //result += sampleDirection;
    }
    return result * oneOverPi * coveredSolidAngle;
}

void main() {
	vec2 uv = vertexOutput.uv;

	vec3 position = texture(gPositionHalfRes, uv).rgb;
	vec3 normal = texture(gNormalHalfRes, uv).rgb;
    vec3 diffuseBrdf = texture(brdfHalfRes, uv).rgb;

	Ray charles = createRay(camera.position, position);
    vec3 LiVolumetric = vec3(0);
	for (int i = 0; i < NUM_LIGHTS; ++i) {
	    Light light = lights[i];
	    LiVolumetric += directionalLightContribution(charles, light);
	}


    vec3 viewSpacePosition = (camera.view * vec4(position, 1)).xyz;
    /*
    mat3 TBN = getCoordinateSystem(normal);
    //TBN = getRotatedCoordinateSystem(normal, uv);

    float ssdoMipLevel = 0;

    vec3 LiSSDO = directionalOcclusion(position, viewSpacePosition, TBN, normal, skybox, ssdoMipLevel);
    vec3 ssdoResult = LiSSDO * diffuseBrdf;
    */

	fragColor = vec4(max(vec3(0), vec3(
	    //ssdoResult,
	    //ssdoResult + LiVolumetric * 4)),
	    LiVolumetric * 4)),
	    //ssdoResult + LiVolumetric,
	    // We pack the fragment's depth in here, so we can use it when upsampling to full resolution.
	    viewSpacePosition.z);
}

